from zeta.zeta import zeta
print(zeta)


#nn
import zeta.nn
from zeta.nn.architecture.transformer import FeedForward
from zeta.nn.modules.layernorm import LayerNorm

#models
import zeta.models


#utils
import zeta.utils


#training
import zeta.training

#tokenizers
import zeta.tokenizers

#rl
import zeta.rl

#optim
import zeta.optim

#ops
import zeta.ops